/*
  This method processes 2x2 block per thread. Because max(sin + cos) = sqrt(2),
  we need only 4 point to approximate whole block. We find minimum among the 
  block pixels, download 4 points starting from minimum and store in local 
  array. Then, we compute linear interpolation for each item.
  
  On NVIDIA it is just slower 80ms instead of 47ms. On AMD the local arrays
  are stored in global memory and everything is extremely slow.
*/

#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;


    
    const int slice_width = 2 * get_global_size(0);

    float res[4] = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = floor(all.z + bx * all.x - by * all.y + all.w);

#pragma unroll 12
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);


#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float4 all = - c_all[proje + i];
            float v[4];
            float h[4];
//            float ch[4];
            int ih[4];

            //float subh = (-sidx * all.x - sidy * all.y) - all.w;
            h[0] = mad(49.f, i, mad(sidx, all.x, mad(sidy, all.y, all.w)));
            h[1] = h[0] - all.x;
            h[2] = h[0] + all.y;
            h[3] = h[1] + all.y;
            
#pragma unroll 4
            for (int j = 0; j < 4; j++) {
                ih[j] = floor(h[j]);
            }

/*
#pragma unroll 4
            for (int j = 0; j < 4; j++)
                res[j] += cache[ih[j]];
*/

            int minh = floor(min(min(h[0], h[1]), min(h[2], h[3])));
            
#pragma unroll 4
            for (int j = 0; j < 4; j++)
                v[j] = cache[minh + j];


#pragma unroll 4
            for (int j = 0; j < 4; j++) {
                float diff = h[j] - ih[j];
                int pos = ih[j] - minh;
                res[j] += v[pos] + diff * (v[pos + 1] - v[pos]);
            }
        
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy    ) + idx] = res[0];
    slice[slice_width * (idy    ) + idx + 1] = res[1];
    slice[slice_width * (idy + 1) + idx    ] = res[2];
    slice[slice_width * (idy + 1) + idx + 1] = res[3];
}
