/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#define _GNU_SOURCE


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <CL/cl.h>

#include <clFFT.h>

#include <glib.h>

#include "hw_tools.h"
#include "debug.h"
#include "hst_setup.h"
#include "hst_reconstructor.h"
#include "hst_opencl_extensions.h"

#define PARALLEL_PER_GPU 1

#ifdef HW_IGNORE_OLD_HARDWARE
# define SHARED_CONTEXT
#endif /* HW_IGNORE_OLD_HARDWARE */

//#define BLOCKED_MEMCPY


#define BLOCK_SIZE 16		// Fix in kernels as well
#define BP_BATCH_SIZE 16
#define FFT_BATCH_SIZE 64       // Due to FFT routines

#define HST_CHECK_MEMORY 512	// minimal size in megabytes
#define HST_OPENCL_CONST_HACK
#define HST_OPENCL_BUILD_FLAGS NULL 
    //"-cl-denorms-are-zero -cl-mad-enable -cl-no-signed-zeros -cl-unsafe-math-optimizations -cl-finite-math-only" 
    //"-cl-nv-maxrregcount=48"
//#define HST_OPENCL_LOAD_PTX "/tmp/hst_opencl.ptx"
//#define HST_OPENCL_SAVE_PTX "/tmp/hst_opencl.ptx"
#define MAX_PROJECTIONS 4000
#define HST_MAX_NAME_LENGTH 64

#define HST_MAX_OPENCL_PLATFORMS 4
#define HST_MAX_OPENCL_DEVICES 16


#ifdef HST_OPENCL_LOAD_PTX
# undef HST_OPENCL_CONST_HACK
#endif /* HST_OPENCL_LOAD_PTX */

#include "hst.h"
#include "hst_opencl.h"

#include "fast_memcpy.h"

static int blocking = 0;
static int have_opencl11 = 1;

static int amd_platform = 0;
static int nvidia_platform = 0;
static int intel_platform = 0;

static cl_context clctx = NULL;
static cl_command_queue clqueue = NULL;

typedef enum {
    HST_DEVICE_UNKNOWN,
    HST_DEVICE_GPU,
    HST_DEVICE_GT200,
    HST_DEVICE_FERMI,
    HST_DEVICE_KEPLER,
    HST_DEVICE_PASCAL,
    HST_DEVICE_CYPRESS,
    HST_DEVICE_GCN,
    HST_DEVICE_AMD_CPU,
    HST_DEVICE_INTEL_CPU,
    HST_DEVICE_NUMBER
} hst_device_type;


typedef enum {
    HST_KERNEL_FILTER,
    HST_KERNEL_BACKPROJECT,
    HST_KERNEL_SPREAD,
    HST_KERNEL_PACK,
    HST_KERNEL_NORMALIZE,
    HST_KERNEL_NUMBER
} hst_kernel_type;


struct OpenCLContextT {
    /* This is one ugly hack... */
    HSTReconstructorContext hst_ctx;
#ifdef PYHST_MEASURE_TIMINGS
# define OpenCL_CONTEXT_MEMSET_OFFSET (6 * sizeof(GTimer*) + sizeof(GTimer**) + sizeof(size_t))
    size_t n_timers;
    GTimer **timers;
    GTimer *main_timer;
    GTimer *init_timer;
    GTimer *fromgpu_timer;
    GTimer *togpu_timer;
    GTimer *texture_timer;
    GTimer *pre_timer;
#else
# define OpenCL_CONTEXT_MEMSET_OFFSET 0
#endif

    int initialized;

    char                *device_name;
    hst_device_type	arch;
    cl_device_type	device_type;
    cl_device_id        device_id;
    cl_context          clctx;
    cl_command_queue    queues[2];
    cl_command_queue    queue;
    cl_program          program;
    cl_image_format     image_format;       /* TODO: We might drop this */
    cl_kernel           kernel[HST_KERNEL_NUMBER];

    clFFT_Plan          fft_plan;

    cl_mem unified_table;   /* everything in one place */
    cl_mem filter;          /* filter coefficients */
    cl_mem slice;           /* final result buffer */
    cl_mem sinogram;
    cl_mem texture;
    cl_mem buffer;          /* keeps FFT results */
    cl_mem frequencies;     /* keeps FFT results */
    
    cl_mem gpures;
    cl_mem hostres;
    void *hostres_ptr;

    cl_mem hostinput;
    void *hostinput_ptr;

    size_t bp_dim[3];
    size_t pad_dim[3];
    size_t input_dim[3];
    size_t filter_dim[3];
    
    float off_x;
    float off_y;
    
    int ppt;		// point per thread
    int bp_block;
    int extra_wraps;    // multiplier

    int blocked_mode;
    int blocked_fft;
    int manual_depading;
    int pinned_memory;
    
#ifdef PYHST_MEASURE_TIMINGS
    cl_ulong timer_init;
    cl_ulong timer_togpu;
    cl_ulong timer_fromgpu;
    cl_ulong timer_filter;
    cl_ulong timer_texture;
    cl_ulong timer_backprojection;
#endif
};

typedef struct OpenCLContextT OpenCLContext;

#include "hst_opencl_error.h"
#include "hst_opencl_kernel.h"
#include "hst_opencl_timers.h"


static int hst_cuda_configure_limits(OpenCLContext *ctx, HSTSetup *setup)
{
    float *param_s;

    int projection;
    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    float slope_zone = setup->pente_zone;

    float param;
    float axis_position_corr;
    float overlapping, flat_zone;

    param_s = (float*) malloc(2 * num_proj * sizeof(float));
    if (!param_s)
        return ENOMEM;

    for (projection = 0; projection < num_proj; projection++) {
        axis_position_corr = setup->axis_position_corr_s[projection];

        if (2 * axis_position_corr > num_bins) {
            overlapping = num_bins - axis_position_corr;
            param = 1; // pente_zone / prof_fact
        }
        else {
            overlapping = axis_position_corr;
            param = -1; // pente_zone / prof_fact
        }

        if (overlapping <= 0)
            return ERANGE;

        slope_zone = MIN(slope_zone, overlapping);
        flat_zone = overlapping - slope_zone;

        param_s[2*projection] = flat_zone;
        param_s[2*projection + 1] = param * slope_zone;
    }

    free(param_s);

    return 0;
}

/*
 * Implementation of HSTReconstructorContext interface
 */
static HSTReconstructorContext *hst_opencl_create_context(HSTReconstructor *prototype, HSTSetup *setup, int id)
{
    int i;
    OpenCLContext *ctx;

    assert(prototype);
    assert(setup);
    assert(id < prototype->devices);

    ctx = (OpenCLContext*) malloc(sizeof(struct OpenCLContextT));
    if (ctx) {
	memset(ctx, 0, sizeof(struct OpenCLContextT));
	
#ifdef PYHST_MEASURE_TIMINGS
        ctx->n_timers = 6;//sizeof(hst_opencl_timers) / sizeof(void*) - 9;
        ctx->timers = (GTimer**)malloc(ctx->n_timers * sizeof(GTimer*));
        if (!ctx->timers) {
            free(ctx);
            return NULL;
        }
        
        for (i = 0; i < ctx->n_timers; i++) {
            ctx->timers[i] = g_timer_new();
            if (!ctx->timers[i]) break;
            g_timer_stop(ctx->timers[i]);
        }
        
        if (i < ctx->n_timers) {
            for (; i >= 0; i--) {
                g_timer_destroy(ctx->timers[i]);
            }
            free(ctx->timers);
            free(ctx);
            return NULL;
        }
        
        ctx->main_timer = ctx->timers[0];
        ctx->pre_timer = ctx->timers[1];
        ctx->togpu_timer = ctx->timers[2];
        ctx->fromgpu_timer = ctx->timers[3];
        ctx->texture_timer = ctx->timers[4];
        ctx->init_timer = ctx->timers[5];
#endif /* PYHST_MEASURE_TIMINGS */

        ctx->device_id = g_hst_platform.device_ids[id/PARALLEL_PER_GPU];
        hst_reconstructor_init_context((HSTReconstructorContext*) ctx, prototype, setup);
    }
    return (HSTReconstructorContext*) ctx;
}

static int hst_opencl_init_context(HSTReconstructorContext *rctx, HWThread thrd)
{
    int err;
    int i;
    int flags;
    OpenCLContext *ctx;
    HSTSetup *setup;


    cl_kernel kernel;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;

    assert(setup);
    assert(setup->num_projections < MAX_PROJECTIONS);

    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    int num_x = setup->num_x;
    int num_y = setup->num_y;
    int dim_fft = setup->dim_fft;
    int dim_proj;

    int variants[HST_MAX_KERNEL_VARIANTS] = { 0 };

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_TYPE, sizeof(cl_device_type), &ctx->device_type, NULL);
    CL_CHECK_ERROR(err);

    ctx->device_name = (char *) malloc(HST_MAX_NAME_LENGTH + 1);
    if (!ctx->device_name) return -1;
    
    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_NAME, HST_MAX_NAME_LENGTH, ctx->device_name, NULL);
    if (err) snprintf(ctx->device_name, 256, "OpenCL %p", ctx->device_id);

    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_NAME, HST_MAX_NAME_LENGTH, ctx->device_name, NULL);

    ctx->pinned_memory = 0;

    if (nvidia_platform) {
	cl_uint compute_major;
	err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_COMPUTE_CAPABILITY_MAJOR_NV, sizeof(cl_uint), &compute_major, NULL); 
	
	if (err) ctx->arch = HST_DEVICE_FERMI;
	else if (compute_major < 2) ctx->arch = HST_DEVICE_GT200;
	else if (compute_major < 3) ctx->arch = HST_DEVICE_FERMI;
	else if (compute_major < 5) ctx->arch = HST_DEVICE_KEPLER;
	else ctx->arch = HST_DEVICE_PASCAL;
        ctx->blocked_mode = BP_BATCH_SIZE;
        ctx->blocked_fft = FFT_BATCH_SIZE;
        ctx->manual_depading = 1;
    } else if (amd_platform) {
	if (ctx->device_type == CL_DEVICE_TYPE_CPU) {
	    ctx->arch = HST_DEVICE_AMD_CPU;
	} else {
	    ctx->manual_depading = 1;

	    if (strcasestr("Tahiti", ctx->device_name)) ctx->arch = HST_DEVICE_GCN;
	    else if (strcasestr("Hawaii", ctx->device_name)) ctx->arch = HST_DEVICE_GCN;
	    else if (strcasestr("Pitcairn", ctx->device_name)) ctx->arch = HST_DEVICE_GCN;
	    else if (strcasestr("Cape Verde", ctx->device_name)) ctx->arch = HST_DEVICE_GCN;
	    else ctx->arch = HST_DEVICE_CYPRESS;
	}
    } else if (intel_platform) {
	ctx->arch = HST_DEVICE_INTEL_CPU;
    } else ctx->arch = HST_DEVICE_UNKNOWN;
    
    ctx->bp_block = BLOCK_SIZE;

#ifdef HST_MULTI_POINT_RECONSTRUCTION
    switch (ctx->arch) {
        case HST_DEVICE_PASCAL:
            ctx->ppt = 6;
            ctx->bp_block = 8;
            break;
        case HST_DEVICE_KEPLER:
            ctx->ppt = 6;
            ctx->bp_block = 8;
            break;
        case HST_DEVICE_CYPRESS:
            if (setup->oversampling > 1) {
                ctx->ppt = 2; //8;
//                ctx->bp_block = 8;
            } else if (setup->oversampling == 0) {
                ctx->ppt = 4; //6;
            }
            break;
        case HST_DEVICE_GCN:
        case HST_DEVICE_FERMI:
	    ctx->ppt = 2;
	    break;
	default:
	    ctx->ppt = 1;
    }
#else /* HST_MULTI_POINT_RECONSTRUCTION */
    ctx->ppt = 1;
#endif /* HST_MULTI_POINT_RECONSTRUCTION */

    ctx->extra_wraps = 1;


    if (num_proj%2) num_proj++;

    ctx->pad_dim[0] = dim_fft;
    ctx->pad_dim[1] = num_proj/2;
    ctx->input_dim[0] = num_bins;//BLOCK_SIZE * hw_calc_blocks(num_bins, ctx->ppt * BLOCK_SIZE, NULL);;
    ctx->input_dim[1] = num_proj/2;
    ctx->filter_dim[0] = 2 * dim_fft;
    ctx->filter_dim[1] = num_proj/2;
    ctx->bp_dim[0] = BLOCK_SIZE * hw_calc_blocks(num_x, ctx->ppt * BLOCK_SIZE, NULL);
    ctx->bp_dim[1] = BLOCK_SIZE * hw_calc_blocks(num_y, ctx->ppt * BLOCK_SIZE, NULL);

    dim_proj = 2 * BLOCK_SIZE * hw_calc_blocks(num_proj, 2 * BLOCK_SIZE, NULL);

#ifdef SHARED_CONTEXT
    if (clctx) ctx->clctx = clctx;
    else {
#endif /* SHARED_CONTEXT */
	ctx->clctx = clCreateContext(0, 1, &ctx->device_id, NULL, NULL, &err);
        CL_CHECK_ERROR(err);
#ifdef SHARED_CONTEXT
    }
#endif /* SHARED_CONTEXT */
#ifdef PYHST_MEASURE_TIMINGS
    flags = CL_QUEUE_PROFILING_ENABLE;
#else
    flags = 0;
#endif

    ctx->queues[0] = clCreateCommandQueue(ctx->clctx, ctx->device_id, flags, &err);
    CL_CHECK_ERROR(err);

    ctx->queues[1] = clCreateCommandQueue(ctx->clctx, ctx->device_id, flags, &err);
    CL_CHECK_ERROR(err);

    ctx->queue = ctx->queues[0];

    /* Allocate and copy buffers, we don't use immediate copying during buffer
     * creation, since we might copy the data to all devices. */
    cl_event event;

#ifdef HST_OPENCL_CONST_HACK
    if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
    cl_float4 *all = (cl_float4*)malloc(dim_proj * sizeof(cl_float4));
    for (i = 0; i < setup->num_projections; i++) {
	all[i].x = setup->cos_s[i];
	all[i].y = setup->sin_s[i];
	all[i].z = setup->axis_position_corr_s[i];
	all[i].w = floor(ctx->ppt * ctx->bp_block * MIN(MIN(all[i].x, - all[i].y), MIN(0., all[i].x - all[i].y)));
    }
    if (setup->num_projections < dim_proj) memset(all + setup->num_projections, 0, (dim_proj - setup->num_projections) * sizeof(cl_float4));
    ctx->unified_table = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY, dim_proj * sizeof(cl_float4), NULL, &err);
    CL_CHECK_ERROR(err);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->unified_table, CL_TRUE, 0, dim_proj*sizeof(cl_float4), all, 0, NULL, &event));
    free(all);
    hst_measure_time(&ctx->timer_init, ctx->queue, event);
#ifdef HST_OPENCL_CONST_HACK
    }
#endif /* HST_OPENCL_CONST_HACK */

    ctx->filter = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 2*setup->dim_fft*sizeof(float), setup->filter, &err);
    CL_CHECK_ERROR(err);

    ctx->buffer = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, 2*setup->dim_fft*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->slice = clCreateBuffer(ctx->clctx, CL_MEM_WRITE_ONLY, ctx->ppt*ctx->ppt*ctx->bp_dim[0]*ctx->bp_dim[1]*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    
    ctx->sinogram = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, num_bins*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->gpures = clCreateBuffer(ctx->clctx, CL_MEM_WRITE_ONLY, num_x * num_y * sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    
    ctx->hostres = clCreateBuffer(ctx->clctx, CL_MEM_ALLOC_HOST_PTR|CL_MEM_WRITE_ONLY, num_x * num_y * sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->hostres_ptr = clEnqueueMapBuffer(ctx->queue, ctx->hostres, CL_TRUE, CL_MAP_READ, 0, num_x * num_y * sizeof(float), 0, NULL, NULL, &err);

    ctx->hostinput = clCreateBuffer(ctx->clctx, CL_MEM_ALLOC_HOST_PTR|CL_MEM_READ_ONLY, num_bins * num_proj * sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->hostinput_ptr = clEnqueueMapBuffer(ctx->queue, ctx->hostinput, CL_TRUE, CL_MAP_WRITE, 0, num_bins * num_proj * sizeof(float), 0, NULL, NULL, &err);


	// Texture
    if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	cl_image_format volume_format;
	volume_format.image_channel_order = CL_R;
	volume_format.image_channel_data_type = CL_FLOAT;

	    // This is crashing with APPSDK 2.7 

#ifdef  CL_VERSION_1_2
	if (!nvidia_platform) {
	    cl_image_desc image_desc;
	    memset(&image_desc, 0, sizeof(cl_image_desc));
	    image_desc.image_type = CL_MEM_OBJECT_IMAGE2D;
	    image_desc.image_width = num_bins;
	    image_desc.image_height = num_proj;
	    ctx->texture = clCreateImage(ctx->clctx, CL_MEM_READ_ONLY, &volume_format, &image_desc, NULL, &err);
	} else {
#endif /* CL_VERSION_1_2 */
	    if (amd_platform) {
		printf("Warning: we may segfault here on AMD platform... OpenCL 1.2 is highly recommended.\n");
	    }
	    ctx->texture = clCreateImage2D(ctx->clctx, CL_MEM_READ_ONLY, &volume_format, num_bins, num_proj, 0, NULL, &err);
#ifdef  CL_VERSION_1_2
	}
#endif /* CL_VERSION_1_2 */

	CL_CHECK_ERROR(err);
    }

    /* Create FFT plan */
    clFFT_Dim3 dim;
    dim.x = setup->dim_fft; dim.y = 1; dim.z = 1;
    ctx->fft_plan = clFFT_CreatePlan(ctx->clctx, dim, clFFT_1D, clFFT_InterleavedComplexFormat, &err);
    CL_CHECK_ERROR(err);


#ifdef HST_FAST_MODE
    variants[HST_KERNEL_BACKPROJECT] = 1;
#else /* HST_FAST_MODE */
    variants[HST_KERNEL_BACKPROJECT] = setup->oversampling;
#endif /* HST_FAST_MODE */

    hst_opencl_load_kernels(ctx, variants);
    
    /* Set the arguments here, to avoid further calls */


    kernel = ctx->kernel[HST_KERNEL_SPREAD];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(int), &setup->dim_fft));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 3, sizeof(int), &setup->num_bins));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(int), &setup->num_bins));

    kernel = ctx->kernel[HST_KERNEL_FILTER];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->filter));

    kernel = ctx->kernel[HST_KERNEL_PACK];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(int), &setup->dim_fft));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 3, sizeof(int), &setup->num_bins));

    kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];

    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(int), &setup->num_projections));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(int), &setup->num_bins));
    if (ctx->device_type == CL_DEVICE_TYPE_GPU) { 
	CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *) &ctx->texture)); 
    } else { 
	CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *) &ctx->sinogram));
    }
    CL_CHECK_ERROR(clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *) &ctx->slice));

#ifdef HST_OPENCL_CONST_HACK
    if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
    CL_CHECK_ERROR(clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *) &ctx->unified_table));
#ifdef HST_OPENCL_CONST_HACK
    }
#endif /* HST_OPENCL_CONST_HACK */


#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    return 0;
}

static int hst_opencl_configure(HSTReconstructorContext *rctx, HSTChanged changes)
{
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    if (changes & HST_FILTER_CHANGED) {
        cl_event event;
        CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->filter, CL_TRUE, 0, 2*setup->dim_fft*sizeof(float), setup->filter, 0, NULL, &event));
        hst_measure_time(&ctx->timer_init, ctx->queue, event);
    }


    if (changes & HST_LIMITS_CHANGED) {
        if (setup->fai360) {
            /* FIXME: check some obscure limits */
        }
        cl_kernel kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];

        ctx->off_x = setup->offset_x - setup->axis_position - 0.5f;
	ctx->off_y = setup->offset_y - setup->axis_position - 0.5f;
	CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(float), &ctx->off_x));
        CL_CHECK_ERROR(clSetKernelArg(kernel, 3, sizeof(float), &ctx->off_y));
    }
    
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    return 0;
}

static void hst_opencl_free_context(HSTReconstructorContext *rctx)
{
    hst_kernel_type kernel_id;
    OpenCLContext *ctx;
    
    assert(rctx);
    ctx = (OpenCLContext*) rctx;

    if (ctx->initialized) {
        free(ctx->device_name);

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_continue(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */

        clFFT_DestroyPlan(ctx->fft_plan);
        if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	    CL_CHECK_ERROR(clReleaseMemObject(ctx->texture));
	}
#ifdef HST_OPENCL_CONST_HACK
	if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
        CL_CHECK_ERROR(clReleaseMemObject(ctx->unified_table));
#ifdef HST_OPENCL_CONST_HACK
	}
#endif /* HST_OPENCL_CONST_HACK */

        CL_CHECK_ERROR(clReleaseMemObject(ctx->filter));

        CL_CHECK_ERROR(clReleaseMemObject(ctx->hostinput));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->hostres));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->gpures));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->slice));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->sinogram));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->buffer));
        CL_CHECK_ERROR(clReleaseCommandQueue(ctx->queues[1]));
        CL_CHECK_ERROR(clReleaseCommandQueue(ctx->queues[0]));
        for (kernel_id = 0; kernel_id < HST_KERNEL_NUMBER; kernel_id++) {
            CL_CHECK_ERROR(clReleaseKernel(ctx->kernel[kernel_id]));
        }
        CL_CHECK_ERROR(clReleaseProgram(ctx->program));

#ifndef SHARED_CONTEXT
        CL_CHECK_ERROR(clReleaseContext(ctx->clctx));
#endif /* SHARED_CONTEXT */

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->init_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }

    memset(((char*) ctx) + sizeof(HSTReconstructorContext) + OpenCL_CONTEXT_MEMSET_OFFSET, 0, sizeof(OpenCLContext) - sizeof(HSTReconstructorContext) - OpenCL_CONTEXT_MEMSET_OFFSET);
}

static void hst_opencl_destroy_context(HSTReconstructorContext *rctx)
{
    int i;
    OpenCLContext *ctx;

    hst_opencl_free_context(rctx);
    hst_reconstructor_free_context(rctx);

    ctx = (OpenCLContext*) rctx;
#ifdef PYHST_MEASURE_TIMINGS
    if (ctx->timers) {
        for (i = 0; i < ctx->n_timers; i++) {
            if (ctx->timers[i]) {
                g_timer_destroy(ctx->timers[i]);
            }
        }
        free(ctx->timers);
    }
#endif /* PYHST_MEASURE_TIMINGS */
    
    free(rctx);
}

static HSTConstString hst_opencl_get_title(const HSTReconstructorContext *rctx)
{
    assert(rctx);
    
    return ((OpenCLContext *) rctx)->device_name;
}

static int hst_opencl_reconstruct(HSTReconstructorContext *rctx, float* slice, const float *sinogram)
{
    int i, j;
    float *real_slice;
    
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;
    assert(setup);
    assert(slice);

    int num_bins = setup->num_bins;
    int num_proj = setup->num_projections;
    int num_x = setup->num_x;
    int num_y = setup->num_y;
    cl_event event;

    cl_int err = CL_SUCCESS;

    size_t zero_offset[] = {0, 0, 0};
    size_t slice_size[] = {num_x * sizeof(float), num_y, 1};
    size_t projection_region[] = {num_bins, num_proj, 1};
    
    size_t local_work_size[] = { BLOCK_SIZE,  BLOCK_SIZE };

    if (setup->fai360) {
        printf("FAI360 requested\n");
    }

    if (ctx->pinned_memory) {
#ifdef PYHST_MEASURE_TIMINGS
        g_timer_continue(ctx->togpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
        fast_memcpy(ctx->hostinput_ptr, sinogram,  num_bins*num_proj*sizeof(float));
        sinogram = ctx->hostinput_ptr;
#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->togpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }


    if (intel_platform) {	// FFT failing at the moment
#ifdef PYHST_MEASURE_TIMINGS
        g_timer_continue(ctx->togpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
        CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->sinogram, CL_TRUE, 0, num_bins*num_proj*sizeof(float), sinogram, 0, NULL, &event));
        hst_measure_time(&ctx->timer_togpu, ctx->queue, event);
#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->togpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    } else {
#ifdef PYHST_MEASURE_TIMINGS
        g_timer_continue(ctx->pre_timer);
#endif /* PYHST_MEASURE_TIMINGS */

        size_t input_dim[3] = {0};
        size_t pad_dim[3] = {0};
        size_t filter_dim[3] = {0};
        size_t fft_offset[3] = {0};

        size_t batch;
        size_t batch_size;

        int fft_proj = num_proj/2;
        if (num_proj%2) {
            // DS: memset to 0, the last projection!!!
            fft_proj++;
        }

        if ((ctx->blocked_fft)&&(have_opencl11)&&(!blocking)) {
            batch_size = ctx->blocked_fft * BLOCK_SIZE;
        } else {
            batch_size = fft_proj;
        }

        input_dim[0] = ctx->input_dim[0];
        input_dim[1] = batch_size; 
        pad_dim[0] = ctx->pad_dim[0];
        pad_dim[1] = batch_size;
        filter_dim[0] = ctx->filter_dim[0];
        filter_dim[1] = batch_size;

        cl_event spread_events[1 + fft_proj / batch_size];
        cl_event pack_events[1 + fft_proj / batch_size];
        cl_event filter_events[1 + fft_proj / batch_size];
        cl_event fft_events[1 + fft_proj / batch_size];
        cl_event ifft_events[1 + fft_proj / batch_size];
        cl_event mem_events[1 + fft_proj / batch_size];
        
        for (batch = 0, i =  0; batch < fft_proj; batch += batch_size, i++) {
            fft_offset[1] = batch;

            if ((batch + batch_size) > fft_proj) {
                batch_size = fft_proj - batch;
                input_dim[1] = batch_size;
                pad_dim[1] = batch_size;
                filter_dim[1] = batch_size;
            }

#ifdef PYHST_MEASURE_TIMINGS
            g_timer_stop(ctx->pre_timer);
            g_timer_continue(ctx->togpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
            CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queues[i%2], ctx->sinogram, blocking?CL_TRUE:CL_FALSE, 2 * batch * num_bins * sizeof(float), 2 * batch_size * num_bins * sizeof(float), sinogram + 2 * batch * num_bins, 0, NULL, &mem_events[i]));
#ifdef PYHST_MEASURE_TIMINGS
            g_timer_stop(ctx->togpu_timer);
            g_timer_continue(ctx->pre_timer);
#endif /* PYHST_MEASURE_TIMINGS */

                // Because we are re-using the start of ctx->buffer
            if (i) clFinish(ctx->queues[(i-1)%2]);

            CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queues[i%2], ctx->kernel[HST_KERNEL_SPREAD], 2, fft_offset, pad_dim, NULL, 0, NULL, &spread_events[i]));

            CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queues[i%2], ctx->fft_plan, batch_size, clFFT_Forward, ctx->buffer, ctx->buffer, 0, NULL, &fft_events[i]));
            CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queues[i%2], ctx->kernel[HST_KERNEL_FILTER], 2, NULL, filter_dim, NULL, 0, NULL, &filter_events[i]));
            CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queues[i%2], ctx->fft_plan, batch_size, clFFT_Inverse, ctx->buffer, ctx->buffer, 0, NULL, &ifft_events[i]));

            CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queues[i%2], ctx->kernel[HST_KERNEL_PACK], 2, fft_offset, input_dim, NULL, 0, NULL, &pack_events[i]));
            if (blocking) clFinish(ctx->queues[i%2]);
        }

        clFinish(ctx->queues[0]);
        clFinish(ctx->queues[1]);
        

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->pre_timer);

            // FFT library is actually ignore events at the moment
        for (j = 0; j < i; j++) {
            hst_measure_time(&ctx->timer_togpu, NULL, mem_events[j]);
            hst_measure_time(&ctx->timer_filter, NULL, spread_events[j]);
//            hst_measure_time(&ctx->timer_filter, NULL, fft_events[j]);
            hst_measure_time(&ctx->timer_filter, NULL, filter_events[j]);
//            hst_measure_time(&ctx->timer_filter, NULL, ifft_events[j]);
            hst_measure_time(&ctx->timer_filter, NULL, pack_events[j]);
        }
#endif /* PYHST_MEASURE_TIMINGS */
    }



#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->texture_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	clEnqueueCopyBufferToImage(ctx->queue, ctx->sinogram, ctx->texture, 0, zero_offset, projection_region, 0, NULL, &event);
	hst_measure_time(&ctx->timer_texture, ctx->queue, event);
    }

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->texture_timer);
    g_timer_continue(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    /* 5. Back-project */
/*    memset(slice, 0, num_x * sizeof(float) * num_y);
    CL_CHECK_ERROR(clEnqueueWriteBufferRect(ctx->queue, ctx->slice, CL_TRUE, zero_offset, zero_offset, slice_size, ctx->ppt * ctx->bp_dim[0] * sizeof(float), 0, 0, 0, slice, 0, NULL, &event));
    memset(slice, 10, num_x * sizeof(float) * num_y);*/


    if (ctx->pinned_memory) {
        real_slice = slice;
        slice = ctx->hostres_ptr;
    }

    local_work_size[0] = ctx->bp_block;
    local_work_size[1] = ctx->bp_block * ctx->extra_wraps;

    if ((ctx->blocked_mode)&&(have_opencl11)&&(!blocking)) {
        size_t last_batch, batch;
        size_t last_batch_size, batch_size = ctx->ppt * ctx->blocked_mode * ctx->bp_block;
        size_t bp_dim[3] = {0};
        size_t bp_offset[3] = {0};
        cl_event bp_events[1 + num_y / batch_size];
        cl_event mem_events[1 + num_y / batch_size];
        cl_event pad_events[1 + num_y / batch_size];

        bp_dim[0] = ctx->bp_dim[0];
        bp_dim[1] = ctx->blocked_mode * ctx->bp_block * ctx->extra_wraps;
        slice_size[1] = batch_size;

        for (batch = 0, i = 0; batch < num_y; batch += batch_size, i++) {
            bp_offset[1] = batch / ctx->ppt;
            zero_offset[1] = batch;

            if ((batch + batch_size) > num_y) {
                batch_size = num_y - batch;
                bp_dim[1] = ctx->bp_block * hw_calc_blocks(batch_size, ctx->ppt * ctx->bp_block, NULL) * ctx->extra_wraps;
                slice_size[1] = batch_size;
            }

            CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queues[i%2], ctx->kernel[HST_KERNEL_BACKPROJECT], 2, bp_offset, bp_dim, local_work_size, 0, NULL, &bp_events[i]));

            if (ctx->manual_depading) {
                CL_CHECK_ERROR(clEnqueueCopyBufferRect(ctx->queues[i%2], ctx->slice, ctx->gpures, zero_offset, zero_offset, slice_size, ctx->ppt * ctx->bp_dim[0] * sizeof(float), 0, 0, 0, 0, NULL, &pad_events[i]));
                CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queues[i%2], ctx->gpures, CL_FALSE, batch * num_x * sizeof(float), batch_size * num_x * sizeof(float), slice + batch * num_x, 0, NULL, &mem_events[i]));

#ifdef BLOCKED_MEMCPY
                if (ctx->pinned_memory) {
                    if (i) {
                        clFinish(ctx->queues[(i-1)%2]);
                        memcpy(real_slice + last_batch * num_x, slice + last_batch * num_x, last_batch_size * num_x * sizeof(float));
                    }
                    last_batch = batch;
                    last_batch_size = batch_size;
                }
#endif /* BLOCKED_MEMCPY */

            } else {
                CL_CHECK_ERROR(clEnqueueReadBufferRect(ctx->queues[i%2], ctx->slice, CL_FALSE, zero_offset, zero_offset, slice_size, ctx->ppt * ctx->bp_dim[0] * sizeof(float), 0, 0, 0, slice, 0, NULL, &mem_events[i]));
            }
        }

        clFinish(ctx->queues[0]);
        clFinish(ctx->queues[1]);


        if (ctx->pinned_memory) {
#ifdef  BLOCKED_MEMCPY
            memcpy(real_slice + last_batch * num_x, slice + last_batch * num_x, last_batch_size * num_x * sizeof(float));
#else /*  BLOCKED_MEMCPY */
            fast_memcpy(real_slice, slice, num_x * num_y * sizeof(float));
#endif /* BLOCKED_MEMCPY */
        }

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->main_timer);

        for (j = 0; j < i; j++) {
            hst_measure_time(&ctx->timer_backprojection, NULL, bp_events[j]);
            hst_measure_time(&ctx->timer_fromgpu, NULL, mem_events[j]);
            hst_measure_time(&ctx->timer_fromgpu, NULL, pad_events[j]);
        }
#endif /* PYHST_MEASURE_TIMINGS */
    } else {
        size_t bp_dim[3] = {0};
        bp_dim[0] = ctx->bp_dim[0];
        bp_dim[1] = ctx->bp_dim[1] * ctx->extra_wraps;

        CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, ctx->kernel[HST_KERNEL_BACKPROJECT], 2, NULL, bp_dim, local_work_size, 0, NULL, &event));
        hst_measure_time(&ctx->timer_backprojection, ctx->queue, event);

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->main_timer);
        g_timer_continue(ctx->fromgpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */

        if (have_opencl11) {
            if (ctx->manual_depading) {
	        CL_CHECK_ERROR(clEnqueueCopyBufferRect(ctx->queue, ctx->slice, ctx->gpures, zero_offset, zero_offset, slice_size, ctx->ppt * ctx->bp_dim[0] * sizeof(float), 0, 0, 0, 0, NULL, &event));
                hst_measure_time(&ctx->timer_fromgpu, ctx->queue, event);
	        CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queue, ctx->gpures, CL_TRUE, 0, num_x*num_y*sizeof(float), slice, 0, NULL, &event));
                hst_measure_time(&ctx->timer_fromgpu, NULL, event);
//                CL_CHECK_ERROR(clEnqueueCopyBuffer(ctx->queue, ctx->gpures, ctx->hostres, 0, 0, num_x*num_y*sizeof(float), 0, NULL, &event));
//                hst_measure_time(&ctx->timer_fromgpu, ctx->queue, event);
            } else {
	        CL_CHECK_ERROR(clEnqueueReadBufferRect(ctx->queue, ctx->slice, CL_TRUE, zero_offset, zero_offset, slice_size, ctx->ppt * ctx->bp_dim[0] * sizeof(float), 0, 0, 0, slice, 0, NULL, &event));
                hst_measure_time(&ctx->timer_fromgpu, NULL, event);
            }
        } else {
            cl_event mem_events[num_y];
	    for (i = 0; i < num_y; i++) {
	        CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queue, ctx->slice, CL_FALSE, i * ctx->ppt * ctx->bp_dim[0] * sizeof(float), num_x * sizeof(float), slice + i * num_x, 0, NULL, &mem_events[i]));
	    }
	    clFinish(ctx->queue);
#ifdef PYHST_MEASURE_TIMINGS
            for (j = 0; j < num_y; j++) {
                hst_measure_time(&ctx->timer_fromgpu, NULL, mem_events[j]);
            }
#endif /* PYHST_MEASURE_TIMINGS */
        }

        if (ctx->pinned_memory) {
            fast_memcpy(real_slice, slice, num_x * num_y * sizeof(float));
        }

#ifdef PYHST_MEASURE_TIMINGS
        g_timer_stop(ctx->fromgpu_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }


    /* 6. Put result back into slice buffer */




    return 0;   /* XXX: What does the return value do? */
}

static HSTReconstructor hst_opencl_info = {
    0,
    hst_opencl_get_title,
    hst_opencl_create_context,
    hst_opencl_init_context,
    hst_opencl_free_context,
    hst_opencl_destroy_context,
    hst_opencl_configure,
    NULL,
    NULL,   /* no pre-processing */
    hst_opencl_reconstruct,
    hst_reconstructor_postprocess_slice,
    NULL,
    hst_opencl_get_timers
};

/*
 * Implementation of public interface
 */
HSTReconstructor *hst_opencl_init(int flags)
{
    cl_int err;
    cl_uint max = 0;
    cl_platform_id selected_platform;
    cl_uint i, num_platforms, num_devices;
    cl_platform_id platforms[HST_MAX_OPENCL_PLATFORMS];
    
    char name[128], version[128];
    char *stmp;
    
    CL_CHECK_ERROR(clGetPlatformIDs(HST_MAX_OPENCL_PLATFORMS, platforms, &num_platforms));

    if (ds_log_check_level(G_LOG_LEVEL_INFO)) {
	cl_uint num_cpu = 0, num_gpu = 0, num_accell = 0;

	pyhst_info("OpenCL Platforms: %u", num_platforms);
	for (i = 0; i < num_platforms; i++) {
    	    err = clGetPlatformInfo (platforms[i], CL_PLATFORM_NAME, sizeof(name), name, NULL);
	    if (err == CL_SUCCESS) err = clGetPlatformInfo (platforms[i], CL_PLATFORM_VERSION, sizeof(version), version, NULL);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_CPU, 0, NULL, &num_cpu);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &num_gpu);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_ACCELERATOR, 0, NULL, &num_accell);
	    if (err == CL_SUCCESS) {
		cl_uint j, units, clock;
		cl_ulong mem;
		cl_device_id devices[HST_MAX_OPENCL_DEVICES]; 

		pyhst_info(" Platform %u: %s, %s (CPU: %u, GPU: %u, Other: %u)", i, name, version, num_cpu, num_gpu, num_accell );

		err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_ALL, HST_MAX_OPENCL_DEVICES, devices, &num_devices);
		for (j = 0; j < num_devices; ++j) {
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_NAME, sizeof(name), name, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DRIVER_VERSION, sizeof(version), version, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(units), &units, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(clock), &clock, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(mem), &mem, NULL);
    		    if (err == CL_SUCCESS) pyhst_debug(" * Device %u: %s v. %s: %u SM, %u MHZ, %lu MB", j, name, version, units, clock, mem/1024/1024);
		    else pyhst_debug(" * Device %u: Failed", i);
		}
	    } else pyhst_info(" Platform %u: Failed", i);
	}
    }

	// Looking for GPU-enabled platform first
    if (flags&HST_RECONSTRUCTOR_USE_GPU) {
	    // FIXME: We should exclude slow devices (check memory, capabilities, etc.)
	for (i = 0; i < num_platforms; i++) {
	    err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &num_devices);
	    if ((err == CL_SUCCESS)&&(num_devices > max)) {
		max = num_devices;
		selected_platform = platforms[i];
	    }
	}
    }

    if (!max) {
#ifdef HW_FALLBACK_CPU
	flags|=HST_RECONSTRUCTOR_USE_CPU;
#endif /* HW_FALLBACK_CPU */

	if (flags&HST_RECONSTRUCTOR_USE_CPU) {
	    for (i = 0; i < num_platforms; i++) {
		err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_CPU, 0, NULL, &num_devices);
		if ((err == CL_SUCCESS)&&(num_devices > 0)) {
		    max = num_devices;
		    selected_platform = platforms[i];
		    break;
		}
	    }
	}
    }
    
    if (max) {
        err = clGetPlatformInfo (selected_platform, CL_PLATFORM_VENDOR, sizeof(name), name, NULL);
        if (err ==  CL_SUCCESS) {
    	    char *found;

	    pyhst_warning("OpenCL Platform: %s", name);
	    
	    found = strstr(name, "ATI");
    	    if (!found) found = strstr(name, "AMD");
    	    if (found) {
    		switch (found[3]) {
    		    case ' ':
    		    case '\t':
    		    case 0:
    			amd_platform = 1;
    		}
    	    } else {
    		found = strcasestr(name, "Advanced Micro Devices");
    		if (found) amd_platform = 1;
    	    }
	    
	    found = strcasestr(name, "Intel");
	    if (found) intel_platform = 1;

	    found = strcasestr(name, "NVIDIA");
	    if (found) nvidia_platform = 1;
    	}

/*	
	if (nvidia_platform) {
	    have_opencl11 = 0; 	// There is a beta driver, but it is incompatible with CUDA 4.0. Very nice of them!
	}
*/
    
	g_hst_platform.platform_id = selected_platform;
	clGetDeviceIDs(g_hst_platform.platform_id,
                ((flags&HST_RECONSTRUCTOR_USE_CPU)?CL_DEVICE_TYPE_CPU:0)|((flags&HST_RECONSTRUCTOR_USE_GPU)?CL_DEVICE_TYPE_GPU:0),
                HST_MAX_OPENCL_DEVICES,
                g_hst_platform.device_ids,
                &(g_hst_platform.num_devices));
	

	int move = 0;
	int have_gpu = 0;
	int ban_next = 0;
	for (i = 0; i < g_hst_platform.num_devices; i++) {
	    cl_device_type device_type;
	    cl_bool device_avail;
	    cl_ulong device_mem;
		
	    if ((amd_platform)&&(flags&HST_RECONSTRUCTOR_USE_GPU)) {
		err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_TYPE, sizeof(cl_device_type), &device_type, NULL);
		CL_CHECK_ERROR(err);
	    }

	    err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_AVAILABLE, sizeof(cl_bool), &device_avail, NULL);
	    CL_CHECK_ERROR(err);

#ifdef HST_CHECK_MEMORY
	    err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &device_mem, NULL);
	    CL_CHECK_ERROR(err);
#endif /* HST_CHECK_MEMORY */

	    if (move) {
		g_hst_platform.device_ids[i - move] =  g_hst_platform.device_ids[i];
	    }
		
	    if (!device_avail) {
	        ++move;
	        ban_next = 0;
	    }
#ifdef HST_CHECK_MEMORY
	    else if (device_mem < HST_CHECK_MEMORY) {
	        ++move;
	        ban_next = 0;
	    }
#endif /* HST_CHECK_MEMORY */
	    else if ((amd_platform)&&(device_type == CL_DEVICE_TYPE_GPU)) {
	        if (ban_next) ban_next = 0;
	        else {
	            struct { cl_uint type; cl_char unused[17]; cl_char bus; cl_char device; cl_char function; } pcie;
		    err = clGetDeviceInfo(g_hst_platform.device_ids[i], 0x4037 /* CL_DEVICE_TOPOLOGY_AMD */, sizeof(pcie), &pcie, NULL);
		    if (err == CL_SUCCESS) {
		        FILE *ftmp;
		        char stmp[128];
		        unsigned long vga_class;
		        
		        sprintf(stmp, "/sys/module/fglrx/drivers/pci:fglrx_pci/0000:%02i:%02i.%i/class", pcie.bus, pcie.device, pcie.function);
		        ftmp = fopen(stmp, "r");
		        if (ftmp) {
		            fscanf(ftmp, "0x%lx", &vga_class);
		            if (vga_class != 0x30000) ++move;
		            fclose(ftmp);
		        }
		        
		    }
		}
		ban_next = 0;
/*		
		if (have_gpu) ++move;
		else have_gpu = 1;
*/
	    } else {
	        ban_next = 0;
	    }

	}
	
	g_hst_platform.num_devices -= move;
    } else {
	g_hst_platform.num_devices = 0;
    }

    
    stmp = getenv("CUDA_LAUNCH_BLOCKING");
    if ((stmp)&&(stmp[0])&&(stmp[0] != '0')) blocking = 1;

    if (!nvidia_platform) {
	stmp = getenv("CUDA_VISIBLE_DEVICES");
	if ((stmp)&&(stmp[0])) {
    	    i = 0;
    	    int id;
    	    char *list = malloc(strlen(stmp) + 1);
    	    if (!list) return NULL;
        
    	    strcpy(list, stmp);
        
    	    stmp = strchr(list, ',');
    	    while (stmp) {
        	*stmp = 0;
        	id = atoi(list);
        	if (id >= g_hst_platform.num_devices) return NULL;
        	g_hst_platform.device_ids[i++] = g_hst_platform.device_ids[id];
            
        	list = stmp + 1;
    	    }
    	    id = atoi(list);
    	    if (id >= g_hst_platform.num_devices) return NULL;
    	    g_hst_platform.device_ids[i++] = g_hst_platform.device_ids[id];
    	    g_hst_platform.num_devices = i;

    	    free(list);
	}
    }

    pyhst_info("OpenCL Devices: %u\n", g_hst_platform.num_devices);
    

#ifdef SHARED_CONTEXT
	// clFFT does not work with shared context on AMD platform
    if ((!amd_platform)&&(g_hst_platform.num_devices > 0)) {
        int max_pos = -1;
        cl_ulong max_mem = 0;
        char device_name[HST_MAX_NAME_LENGTH + 1];
        
        for (i = 0; i < g_hst_platform.num_devices; i++) {
	    cl_ulong device_mem;

	    err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &device_mem, NULL);
	    if (err != CL_SUCCESS) continue;
	    
	    if (device_mem > max_mem) {
	        if (amd_platform) {
                    if (clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_NAME, HST_MAX_NAME_LENGTH, device_name, NULL)) continue;
                    if (!strcasecmp(device_name, "Cypress")) continue;
                }
	        max_pos = i;
	        max_mem = device_mem;
	    }
	}
        
        if (max_pos >= 0) {
            clctx = clCreateContext(NULL, g_hst_platform.num_devices, g_hst_platform.device_ids, NULL, NULL, &err);
                // find device with maximum memory
            if (err == CL_SUCCESS) clqueue = clCreateCommandQueue(clctx, g_hst_platform.device_ids[max_pos], 0, &err);
        }
    }
#endif /* SHARED_CONTEXT */

    hst_opencl_info.devices = g_hst_platform.num_devices * PARALLEL_PER_GPU;
    return &hst_opencl_info;
}

void hst_opencl_free() {
    if (clctx) {
        if (clqueue) {
            clReleaseCommandQueue(clqueue);
            clqueue = NULL;
        }
        clReleaseContext(clctx);
        clctx = NULL;
    }
}

#define MAX_HOST_BUFFERS 16
static int hst_opencl_host_buffers = 0;
static void* hst_opencl_host_memory[MAX_HOST_BUFFERS];

void *hst_opencl_host_malloc(size_t size, HSTPinnedAccess access) {
    cl_int err;
    void *tmp;
    
    if (hst_opencl_host_buffers == MAX_HOST_BUFFERS) {
	printf("Too many hosts buffers allocated, something is wrong...\n");
	return NULL;
    }

//    printf("trying: %zu MB\n", size/1024/1024);
    if ((clctx)&&(clqueue)) {
        cl_int flags = 0;
        if (access == HST_PINNED_R) flags = CL_MEM_READ_ONLY;
        else if (access == HST_PINNED_W) flags = CL_MEM_WRITE_ONLY;
        else flags = CL_MEM_READ_WRITE;
        
        cl_mem clres = clCreateBuffer(clctx, CL_MEM_ALLOC_HOST_PTR|flags, size, NULL, &err);
        if (err != CL_SUCCESS) {
            printf("OpenCL failed to allocate %zu MB host memory, error: %i\n", size/1024/1024, err);
	    CL_CHECK_ERROR(err);
            return malloc(size);
        }

        cl_map_flags map_flags = 0;

        if (access&HST_PINNED_R) {
            map_flags |= CL_MAP_WRITE;
        }

        if (access&HST_PINNED_W) {
            map_flags |= CL_MAP_READ;
        }
        tmp = clEnqueueMapBuffer(clqueue, clres, CL_TRUE, map_flags, 0, size, 0, NULL, NULL, &err);
        if (err != CL_SUCCESS) {
            printf("OpenCL failed to map %zu MB host memory, error: %i\n", size/1024/1024, err);
	    CL_CHECK_ERROR(err);
            clReleaseMemObject(clres);
            return malloc(size);
        }

        hst_opencl_host_memory[hst_opencl_host_buffers++] = tmp;
    } else {
        tmp = malloc(size);
        if (tmp) {
            if (mlock(tmp, size)) {
                free(tmp);
                return NULL;
            }
        }
    }

    return tmp;

}

void hst_opencl_host_free(void *ptr) {
    cl_int err;

    if ((clctx)&&(clqueue)) {
	int i;
	for (i = 0; i < hst_opencl_host_buffers; i++) {
	    if (hst_opencl_host_memory[i] == ptr) break;
    	}
    	if (i < hst_opencl_host_buffers) {
	    // Hell with it
	    //clReleaseMemObject(clres)
	} else {
	    free(ptr);
	}
    } else {
        // munlock ?
	free(ptr);
    }
}
